declare module "electron-auto-updater/out/src/util/httpRequest" {
  /// <reference types="node" />
  import { ClientRequest } from "http"

  export interface DownloadOptions {
    skipDirCreation?: boolean
    sha2?: string
  }

  export function download(url: string, destination: string, options?: DownloadOptions | null): Promise<string>

  export function addTimeOutHandler(request: ClientRequest, callback: (error: Error) => void): void
}

declare module "electron-auto-updater/out/src/publish/restApiRequest" {
  /// <reference types="node" />
  import { RequestOptions } from "https"
  import { IncomingMessage, ClientRequest } from "http"
  import { Url } from "url"

  export function githubRequest<T>(path: string, token: string | null, data?: {
    [name: string]: any
  } | null, method?: string): Promise<T>

  export function bintrayRequest<T>(path: string, auth: string | null, data?: {
    [name: string]: any
  } | null, method?: string): Promise<T>

  export function request<T>(url: Url, token?: string | null, data?: {
    [name: string]: any
  } | null, method?: string): Promise<T>

  export function doApiRequest<T>(options: RequestOptions, token: string | null, requestProcessor: (request: ClientRequest, reject: (error: Error) => void) => void, redirectCount?: number): Promise<T>

  export class HttpError extends Error {
    response: IncomingMessage
    description: any
    constructor(response: IncomingMessage, description?: any)
  }
}

declare module "electron-auto-updater/out/src/options/publishOptions" {
  export type PublishProvider = "github" | "bintray" | "generic"
  export type Publish = string | Array<string> | PublishConfiguration | GithubOptions | BintrayOptions | GenericServerOptions | Array<PublishConfiguration> | Array<GithubOptions> | Array<GenericServerOptions> | Array<BintrayOptions> | null

  export interface PublishConfiguration {
    provider: PublishProvider
    owner?: string
    token?: string
  }

  export interface GenericServerOptions extends PublishConfiguration {
    url: string
    /**
    The channel. Defaults to `latest`.
     */
    channel?: string | null
  }

  export interface VersionInfo {
    readonly version: string
  }

  export interface UpdateInfo extends VersionInfo {
    readonly path: string
    readonly sha2: string
  }

  export interface GithubOptions extends PublishConfiguration {
    repo?: string
    vPrefixedTagName?: boolean
  }

  export interface BintrayOptions extends PublishConfiguration {
    package?: string
    repo?: string
    user?: string
  }
}

declare module "electron-auto-updater/out/src/publish/bintray" {
  import { BintrayOptions } from "electron-auto-updater/out/src/options/publishOptions"

  export interface Version {
    readonly name: string
    readonly package: string
  }

  export interface File {
    name: string
    path: string
    sha1: string
    sha256: string
  }

  export class BintrayClient {
    private readonly basePath
    readonly auth: string | null
    readonly repo: string
    readonly owner: string
    private readonly user
    readonly packageName: string
    constructor(options: BintrayOptions, apiKey?: string | null)
    getVersion(version: string): Promise<Version>
    getVersionFiles(version: string): Promise<Array<File>>
    createVersion(version: string): Promise<any>
    deleteVersion(version: string): Promise<any>
  }
}

declare module "electron-auto-updater/out/nsis-auto-updater/src/api" {
  import { VersionInfo } from "electron-auto-updater/out/src/options/publishOptions"

  export interface FileInfo {
    name: string
    url: string
    sha2?: string
  }

  export interface Provider<T extends VersionInfo> {
    getLatestVersion(): Promise<T>
    getUpdateFile(versionInfo: T): Promise<FileInfo>
  }

  export interface UpdateCheckResult {
    readonly versionInfo: VersionInfo
    readonly fileInfo?: FileInfo
    readonly downloadPromise?: Promise<any> | null
  }
}

declare module "electron-auto-updater/out/nsis-auto-updater/src/BintrayProvider" {
  import { Provider, FileInfo } from "electron-auto-updater/out/nsis-auto-updater/src/api"
  import { BintrayOptions, VersionInfo } from "electron-auto-updater/out/src/options/publishOptions"

  export class BintrayProvider implements Provider<VersionInfo> {
    private client
    constructor(configuration: BintrayOptions)
    getLatestVersion(): Promise<VersionInfo>
    getUpdateFile(versionInfo: VersionInfo): Promise<FileInfo>
  }
}

declare module "electron-auto-updater/out/nsis-auto-updater/src/GenericProvider" {
  import { Provider, FileInfo } from "electron-auto-updater/out/nsis-auto-updater/src/api"
  import { GenericServerOptions, UpdateInfo } from "electron-auto-updater/out/src/options/publishOptions"

  export class GenericProvider implements Provider<UpdateInfo> {
    private readonly configuration
    private readonly baseUrl
    private readonly channel
    constructor(configuration: GenericServerOptions)
    getLatestVersion(): Promise<UpdateInfo>
    getUpdateFile(versionInfo: UpdateInfo): Promise<FileInfo>
  }

  export function validateUpdateInfo(info: UpdateInfo): void
}

declare module "electron-auto-updater/out/nsis-auto-updater/src/GitHubProvider" {
  import { Provider, FileInfo } from "electron-auto-updater/out/nsis-auto-updater/src/api"
  import { VersionInfo, GithubOptions, UpdateInfo } from "electron-auto-updater/out/src/options/publishOptions"

  export class GitHubProvider implements Provider<VersionInfo> {
    private readonly options
    constructor(options: GithubOptions)
    getLatestVersion(): Promise<UpdateInfo>
    private getBasePath()
    getUpdateFile(versionInfo: UpdateInfo): Promise<FileInfo>
  }
}

declare module "electron-auto-updater" {
  export const autoUpdater: any
}

declare module "electron-auto-updater/out/nsis-auto-updater/src/NsisUpdater" {
  /// <reference types="node" />
  import { EventEmitter } from "events"
  import { UpdateCheckResult } from "electron-auto-updater/out/nsis-auto-updater/src/api"
  import { BintrayProvider } from "electron-auto-updater/out/nsis-auto-updater/src/BintrayProvider"
  import { BintrayOptions, PublishConfiguration, GithubOptions, GenericServerOptions } from "electron-auto-updater/out/src/options/publishOptions"
  import { GenericProvider } from "electron-auto-updater/out/nsis-auto-updater/src/GenericProvider"
  import { GitHubProvider } from "electron-auto-updater/out/nsis-auto-updater/src/GitHubProvider"

  export class NsisUpdater extends EventEmitter {
    private setupPath
    private updateAvailable
    private quitAndInstallCalled
    private clientPromise
    private readonly app
    private quitHandlerAdded
    constructor(options?: PublishConfiguration | BintrayOptions | GithubOptions)
    getFeedURL(): string | null | undefined
    setFeedURL(value: string | PublishConfiguration | BintrayOptions | GithubOptions | GenericServerOptions): void
    checkForUpdates(): Promise<UpdateCheckResult>
    private doCheckForUpdates()
    private addQuitHandler()
    quitAndInstall(): void
    private install(isSilent)
    loadUpdateConfig(): Promise<BintrayProvider | GenericProvider | GitHubProvider>
  }
}

